#ifndef PCS_STATE_H
#define PCS_STATE_H

//---------------------------------------------------------------------
//    ___
//   /\__\    This file is part of MUSE    <http://www.muse-tools.org/>
//  /::L_L_
// /:/L:\__\  Miami   University  Simulation  Environment    (MUSE)  is
// \/_/:/  /  free software: you can  redistribute it and/or  modify it
//   /:/  /   under the terms of the GNU  General Public License  (GPL)
//   \/__/    as published  by  the   Free  Software Foundation, either
//            version 3 (GPL v3), or  (at your option) a later version.
//    ___
//   /\__\    MUSE  is distributed in the hope that it will  be useful,
//  /:/ _/_   but   WITHOUT  ANY  WARRANTY;  without  even  the IMPLIED
// /:/_/\__\  WARRANTY of  MERCHANTABILITY  or FITNESS FOR A PARTICULAR
// \:\/:/  /  PURPOSE.
//  \::/  /
//   \/__/    Miami University  and  the MUSE  development team make no
//            representations  or  warranties  about the suitability of
//    ___     the software,  either  express  or implied, including but
//   /\  \    not limited to the implied warranties of merchantability,
//  /::\  \   fitness  for a  particular  purpose, or non-infringement.
// /\:\:\__\  Miami  University and  its affiliates shall not be liable
// \:\:\/__/  for any damages  suffered by the  licensee as a result of
//  \::/  /   using, modifying,  or distributing  this software  or its
//   \/__/    derivatives.
//
//    ___     By using or  copying  this  Software,  Licensee  agree to
//   /\  \    abide  by the intellectual  property laws,  and all other
//  /::\  \   applicable  laws of  the U.S.,  and the terms of the  GNU
// /::\:\__\  General  Public  License  (version 3).  You  should  have
// \:\:\/  /  received a  copy of the  GNU General Public License along
//  \:\/  /   with MUSE.  If not,  you may  download  copies  of GPL V3
//   \/__/    from <http://www.gnu.org/licenses/>.
//
//---------------------------------------------------------------------

/* 
 * File:   PCS_State.h
 * Author: Julius Higiro
 *
 * Created on September 1, 2016, 3:05 PM
 */

#include "State.h"

// A global shortcut for unsigned int
using uint = unsigned int;

/** The state associated with the PCSAgent or "Cell".  PCSAgent models
    a PCS "Cell". A Cell represents a cellular receiver/transmitter
    (or a cell tower) that has some fixed number of "channels"
    allocated to it and is the central entity (LP) object type for the
    simulation. A "channel" is a wireless channel via which a
    "portable" (or cellular device) can send/receive information from
    a Cell.  Portables are modeled using different values in the
    PCSEvent.
 */
class PCS_State : public muse::State {
public:
    /** Overrides the default implementation in base cass to return a
        clone of this state.

        The implementation of this method is rather simple and uses
        the copy constructor to create a clone.  Keep in mind this a
        shallow copy.  However, it works here only because there are no
        complex pointers/objects in the state but just primitive data
        types.  So the default copy-constructor  generated by the
        compiler is sufficient.

        \return A new copy/clone of this object.
    */
    State* getClone() override;

    /** The default constructor for this state.

        It initializes all the instance variables to zeros.
    */
    PCS_State();

    /** Convenience constructor to initialize a subset of state values.

        This is a convenience constructor used to initialize state
        values.

        \param[in] idleChannels During initialization, this value
        inidicates the maximum number of channels that this Cell has.
    */
    PCS_State(uint idleChannels);
    
    inline int getIndex() const {
        return index;
    }

    inline void setIndex(int new_index) {
        index = new_index;
    }

    inline uint getBlockedChannels() const {
        return blockedChannels;
    }

    inline void setBlockedChannels(uint value) {
        blockedChannels = value;
    }

    inline uint getCallAttempts() const {
        return callAttempts;
    }

    inline void setCallAttempts(uint value) {
        callAttempts = value;
    }

    inline uint getHandOffBlocks() const {
        return handOffBlocks;
    }

    inline void setHandOffBlocks(uint value) {
        handOffBlocks = value;
    }

    inline uint getIdleChannels() const {
        return idleChannels;
    }

    inline void setIdleChannels(uint value) {
        idleChannels = value;
    }

    inline uint getBusyLineCalls() const {
        return busyLineCalls;
    }
    
    inline void setBusyLineCalls(uint value) {
        busyLineCalls = value;
    }

    /** Track a move-in event where a portable entered this Cell.

        This method merely increments a counter to keep track of
        move-in operations.
     */
    void trackMoveIn() { moveIns++; }

    /** Track a move-out event where a portable left this Cell.

        This method merely increments a counter to keep track of
        move-out operations.
     */    
    void trackMoveOut() { moveOuts++; }

    /** Convenience method to return a string representation of this
        state for printing final statistics.
    */
    std::string to_string() const;
    
private:
    /** A counter used to track the next cell to which a call should
        be dispatched. This counter enables a deterministic hand-off
        operation.
    */
    int index;

    /** 
     * The number of blocked calls that occur because channels
     * are not available.
     */
    uint blockedChannels;

    /** 
     * The number of calls attempted.
     */
    uint callAttempts;

    /** 
     * The number of blocked calls that occur because channels
     * are not available when engaged PCSEvent/Portables are in transit to 
     * a new destination PCSAgent/Cell. 
     */
    uint handOffBlocks;

    /**
     * The number of channels available on this Cell to be potentially
     * used for a call by a portable.  Whenever a portable receives a
     * call, this value is decremented.  When a call completes or a
     * portable moves, this value is incremented.  This value is
     * initialized with the maximum number of channels that will
     * assigned to this Cell/agent.
     */
    uint idleChannels;

    /** The number of calls that could not be received as the line was
        already busy.  See nextCall() method on how this value is
        incremented.
    */
    uint busyLineCalls;

    /** The number of times a portable moved into this Cell. */
    uint moveIns;

    /** The number of times a portable moved out of this Cell. */    
    uint moveOuts;
};


#endif /* PCS_STATE_H */
